## NVIDIA GPU Operator

###  Hinweise

Für die Nutzung von CUDA in Kubernetes Knoten:
Siehe https://docs.nvidia.com/datacenter/cloud-native/gpu-operator/latest/gpu-driver-configuration.html

Für die Nutzung mit vorinstallierten Treibern und Toolkit:
Siehe https://www.jimangel.io/posts/nvidia-rtx-gpu-kubernetes-setup/

### Installation

- Repository und Values Datei

    ```bash
    helm repo add nvidia https://helm.ngc.nvidia.com/nvidia 
    helm repo update nvidia
    helm search repo -l nvidia/gpu-operator

    cd <path>/containerize/helm-values/NVidia
    # Default Values ausgeben
    helm inspect values nvidia/gpu-operator > gpu-operator-values-default.yaml
    ```

- angepasste values-Datei `gpu-operator-values.yaml` erzeugen 

- Chart einspielen, Empfehlung Namespace `gpu-operator`

    ```bash
    helm upgrade --install --wait gpu-operator \
        -n gpu-operator --create-namespace \
        nvidia/gpu-operator \
        --values gpu-operator-values.yaml
    # check rollout
    kubectl rollout status deployment gpu-operator -n gpu-operator
    helm status gpu-operator -n gpu-operator
    ```

#### Variante Bare-metal/Passthrough mit pre-installed NVIDIA-Treiber

```yaml
driver:
  enabled: false
  usePrecompiled: true
  repository: nvcr.io/nvidia
  version: "550.54.15"  # hier muss die passende Version NVidia-Treiber der Basismaschine eingetragen werden 
  
toolkit:
  enabled: true
  #env: []
  env:
   - name: CONTAINERD_CONFIG
     value: /var/lib/rancher/rke2/agent/etc/containerd/config.toml.tmpl
   - name: CONTAINERD_SOCKET
     value: /run/k3s/containerd/containerd.sock
   - name: CONTAINERD_RUNTIME_CLASS
     value: nvidia
   - name: CONTAINERD_SET_AS_DEFAULT
     value: "true"  
```

#### Variante Bare-metal/Passthrough mit pre-installed NVIDIA-Treiber und -Toolkit

- Toolkit instlallieren 

```bash
# add nvidia-container-toolkit repo to apt sources
curl -fsSL https://nvidia.github.io/libnvidia-container/gpgkey | sudo gpg --dearmor -o /usr/share/keyrings/nvidia-container-toolkit-keyring.gpg \
&& curl -s -L https://nvidia.github.io/libnvidia-container/stable/deb/nvidia-container-toolkit.list | \
sed 's#deb https://#deb [signed-by=/usr/share/keyrings/nvidia-container-toolkit-keyring.gpg] https://#g' | \
sudo tee /etc/apt/sources.list.d/nvidia-container-toolkit.list
sudo apt update

# install container toolkit
sudo apt install -y nvidia-container-toolkit
# options: --dry-run
sudo nvidia-ctk runtime configure --runtime=containerd

reboot
```

```yaml
driver:
  enabled: false
  version: "550.127.05"  # hier muss die passende Version NVidia-Treiber der Basismaschine eingetragen werden 
  
toolkit:
  enabled: false
  env:
   - name: CONTAINERD_CONFIG
     value: /var/lib/rancher/rke2/agent/etc/containerd/config.toml.tmpl
   - name: CONTAINERD_SOCKET
     value: /run/k3s/containerd/containerd.sock
   - name: CONTAINERD_RUNTIME_CLASS
     value: nvidia
   - name: CONTAINERD_SET_AS_DEFAULT
     value: "true"  
```

### Time-Slicing
<https://docs.nvidia.com/datacenter/cloud-native/gpu-operator/latest/gpu-sharing.html>

Ermöglicht es in Kubernetes eine GPU mehrfach zu reservieren. Die Leistung wird dabei automatisch von der GPU aufgeteilt.
D.h., wenn nur ein Pod die GPU nutzt, erhält dieser die volle Leistung, auch wenn er nur einen "Time-Slice" reserviert hat.

Die Einstellung funktioniert über eine Config-Map die manuell im Kubernetes angelegt und dann dem GPU-Operator
bekannt gemacht werden muss.

Config Map hinzufügen
```bash
kubectl create -n gpu-operator -f time-slicing-config-all.yaml
```

Config Map dem GPU-Operator bekannt machen
```bash
kubectl patch clusterpolicies.nvidia.com/cluster-policy \
    -n gpu-operator --type merge \
    -p '{"spec": {"devicePlugin": {"config": {"name": "time-slicing-config-all", "default": "any"}}}}'
```

Danach starten die Pods "gpu-feature-discovery" und "nvidia-device-plugin" neu, und die Flags der Nodes:
* nvidia.com/gpu.replicas
* nvidia.com/gpu.sharing-strategy

werden geupdated.

Der neue Name der GPU endet mit einem `-SHARED` um anzuzeigen, dass es sich um geteilte GPUs handelt. Die kann mit
der Einstellung `renameByDefault` geändert werden.